/*
 * lpcload.c - load firmware into ram of lpc2220 via uart0
 *
 * author: hackbard@hackdaworld.org, rolf.anders@physik.uni-augsburg.de
 *
 * build: make
 * usage: sudo ./lpcload -d /dev/ttyS0 -f firmware.hex [-v]
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <termios.h>

#define VERBOSE			(1<<0)
#define FIRMWARE		(1<<1)

#define TXRX_TYPE_BAUD		0x01
#define TXRX_TYPE_SYNC		0x02
#define TXRX_TYPE_CMD		0x03
#define TXRX_TYPE_DATA		0x04
#define TXRX_TYPE_CKSM		0x05

#define CMD_SUCCESS		"0\r\n"
#define INVALID_COMMAND		"1\r\n"
#define SRC_ADDR_ERROR		"2\r\n"
#define DST_ADDR_ERROR		"3\r\n"
#define SRC_ADDR_NOT_MAPPED	"4\r\n"
#define DST_ADDR_NOT_MAPPED	"5\r\n"
#define COUNT_ERROR		"6\r\n"
#define COMPARE_ERROR		"10\r\n"
#define BUSY			"11\r\n"
#define PARAM_ERROR		"12\r\n"
#define ADDR_ERROR		"13\r\n"
#define ADDR_NOT_MAPPED		"14\r\n"
#define CMD_LOCKED		"15\r\n"
#define INVALID_CODE		"16\r\n"
#define INVALID_BAUD_RATE	"17\r\n"
#define INVALID_STOP_BIT	"18\r\n"

#define CRYSTFREQ		"10000"
#define RAMOFFSET		0x40000200

#define BUFSIZE			128

typedef unsigned char u8;
typedef unsigned short u16;
typedef unsigned int u32;

typedef struct s_lpc {
	int sfd;		/* serial fd */
	char sdev[128];		/* seriel device */
	int fwfd;		/* fimrware fd */
	char fwfile[128];	/* firmware file */
	u8 info;		/* info/mode */
	char freq[8];		/* frequency */
	u32 hoff;		/* start addr of ihex file */
	u32 roff;		/* ram offset of uc */
} t_lpc;

void usage(void) {

	printf("possible argv:\n");
	printf("  -d <serial device>\n");
	printf("  -f <firmware>\n");
	printf("  -c <crystal freq>\n");
	printf("  -r <ram offset>\n");
	printf("  -v\n");

}

int open_serial_device(t_lpc *lpc) {

	struct termios term;

	//memset(&term,0,sizeof(struct termios));

	/* open serial device */

	lpc->sfd=open(lpc->sdev,O_RDWR);
	if(lpc->sfd<0) {
		perror("tts open");
		return lpc->sfd;
	}

	/* configure the serial device */

	tcgetattr(lpc->sfd,&term);

	// input/output baudrate

	cfsetispeed(&term,B9600);
	cfsetospeed(&term,B9600);

	// control options -> 8n1

	term.c_cflag&=~PARENB;	// no parity
	term.c_cflag&=~CSTOPB;	// only 1 stop bit
	term.c_cflag&=~CSIZE;	// no bit mask for data bits
	term.c_cflag|=CS8;	// 8 data bits

	// line options -> raw input
	
	term.c_lflag&=~(ICANON|ECHO|ECHOE|ISIG);

	// input options -> enable flow control
	
	//term.c_iflag&=~(IXON|IXOFF|IXANY|INLCR|ICRNL);
	term.c_iflag&=~(INLCR|ICRNL|IXANY);
	term.c_iflag|=(IXON|IXOFF);
	
	// output options

	term.c_oflag=0;

	// more control options -> timeout / flow control
	
	term.c_cc[VMIN]=0;
	term.c_cc[VTIME]=40;	// 4 second timeout
	term.c_cc[VSTART]=0x11;
	term.c_cc[VSTOP]=0x13;

	tcsetattr(lpc->sfd,TCSANOW,&term);

	return lpc->sfd;
}

int open_firmware(t_lpc *lpc) {

	int ret;
	char buf[BUFSIZE];

	/* open firmware file */

	lpc->fwfd=open(lpc->fwfile,O_RDONLY);

	if(lpc->fwfd<0)
		perror("fw open");

	/* read hex file offset */

	ret=read(lpc->fwfd,buf,7);
	if(buf[0]!=':') {
		printf("fw open: not an intel hex file?\n");
		return -1;
	}
	sscanf(buf+3,"%04x",&(lpc->hoff));
	lseek(lpc->fwfd,0,SEEK_SET);

	return lpc->fwfd;
}

int txrx(t_lpc *lpc,char *buf,int len,u8 type) {

	int ret,cnt;
	int i;

	/* write */

	if(lpc->info&VERBOSE)
		printf("  >> ");
	cnt=0;
	while(len) {
		ret=write(lpc->sfd,buf+cnt,len);
		if(ret<0) {
			perror("txrx write");
			return ret;
		}
		if(lpc->info&VERBOSE)
			for(i=0;i<ret;i++)
				printf("%c",
				       ((buf[cnt+i]>0x19)&(buf[cnt+i]<0x7f))?
				       buf[cnt+i]:'.');
		len-=ret;
		cnt+=ret;
	}
	if(lpc->info&VERBOSE) {
		printf(" | ");
		for(i=0;i<cnt;i++)
			printf("%02x ",buf[i]);
		printf("| (%d)\n",cnt);
	}

	/* cut the echo if not of type auto baud */

	if(type!=TXRX_TYPE_BAUD) {
		while(cnt) {
			ret=read(lpc->sfd,buf,cnt);
			if(ret<0) {
				perror("txrx echo cut");
				return ret;
			}
			cnt-=ret;
		}
	}

	/* return here if type is data */

	if(type==TXRX_TYPE_DATA)
		return cnt;

	/* read */

	ret=read(lpc->sfd,buf,1);
	if(ret<0) {
		perror("txrx read (first byte)");
		return ret;
	}
		
	switch(buf[0]) {
		case 'S':
			cnt=13;
			break;
		case 'O':
			cnt=3;
			break;
		case 'R':
			cnt=7;
			break;
		case '0':
			cnt=2;
			break;
		default:
			printf("txrx read: bad return byte '%02x'\n",buf[0]);
			break;
	}

	ret=1;
	i=cnt;
	while(i) {
		ret=read(lpc->sfd,buf+1+cnt-i,i);
		if(ret<0) {
			perror("txrx read (next bytes)");
			return ret;
		}
		i-=ret;
	}
	if(lpc->info&VERBOSE) {
		printf("  << ");
		for(i=0;i<cnt+1;i++)
			printf("%c",((buf[i]>0x19)&(buf[i]<0x7f))?
			            buf[i]:'.');
		printf(" | ");
		for(i=0;i<cnt+1;i++)
			printf("%02x ",buf[i]);
		printf("| (%d)\n",cnt+1);
	}
	buf[cnt+1]='\0';

	/* check/strip return code if type is cmd */

	if(type==TXRX_TYPE_CMD) {
		ret=strlen(CMD_SUCCESS);
		if(!strncmp(buf,CMD_SUCCESS,ret)) {
			for(i=ret;i<cnt;i++)
				buf[i-ret]=buf[i];
			buf[cnt]='\0';
		}
		else {
			printf("txrx bad return code!\n");
			return -1;
		}
	}

	return cnt;
}

int bl_init(t_lpc *lpc) {

	char buf[BUFSIZE];
	int len;

	/* auto baud sequence */
	buf[0]='?';
	txrx(lpc,buf,1,TXRX_TYPE_BAUD);
	if(strncmp(buf,"Synchronized\r\n",14)) {
		printf("auto baud detection failed\n");
		return -1;
	}

	/* tell bl that we are synchronized (it's allready in buf) */
	txrx(lpc,buf,14,TXRX_TYPE_SYNC);
	if(strncmp(buf,"OK\r\n",4)) {
		printf("sync failed\n");
		return -1;
	}

	/* tell bl the crystal frequency */
	len=strlen(lpc->freq)+2;
	strncpy(buf,lpc->freq,BUFSIZE);
	buf[len-2]='\r';
	buf[len-1]='\n';
	txrx(lpc,buf,len,TXRX_TYPE_SYNC);
	if(strncmp(buf,"OK\r\n",4)) {
		printf("freq set failed\n");
		return -1;
	}

	return 0;
}

int unlock_go(t_lpc *lpc) {

	char buf[BUFSIZE];
	int ret;

	memcpy(buf,"U 23130\r\n",9);
	ret=txrx(lpc,buf,9,TXRX_TYPE_CMD);

	return ret;
}

int go(t_lpc *lpc) {

	char buf[BUFSIZE];
	int ret,len;

	snprintf(buf,BUFSIZE,"G %d A\r\n",lpc->roff);
	len=strlen(buf);
	ret=txrx(lpc,buf,len,TXRX_TYPE_CMD);

	return ret;
}

int uuencode(u8 *in,u8 *out,int len) {

	out[0]=0x20+len;
	out[1]=0x20+((in[0]>>2)&0x3f);
	out[2]=0x20+(((in[0]<<4)|(in[1]>>4))&0x3f);
	out[3]=0x20+(((in[1]<<2)|(in[2]>>6))&0x3f);
	out[4]=0x20+(in[2]&0x3f);

	return 0;
}

int write_to_ram(t_lpc *lpc,char *buf,u32 addr,int len) {

	int lcount;
	u32 checksum;
	char txrxbuf[BUFSIZE];
	int count,bcnt;
	int nlen,slen;
	int i;

	/* check length */
	if(len%4) {
		printf("ram write: not a multiple of 4\n");
		return -1;
	}

	/* make it a multiple of 3 (reason: uuencode) */
	nlen=(!(len%3))?len:((len/3+1)*3);
	if(nlen>BUFSIZE) {
		printf("ram write: too much data\n");
		return -1;
	}
	for(i=len;i<nlen;i++) buf[i]=0;

	/* prepare addr */
	addr+=(lpc->roff-lpc->hoff);

	/* prepare write command */
	if(lpc->info&VERBOSE)
		printf("writing 0x%02x bytes to 0x%08x\n",len,addr);
	snprintf(txrxbuf,BUFSIZE,"W %d %d\r\n",addr,len);
	slen=strlen(txrxbuf);

	/* send command and check return code */
	txrx(lpc,txrxbuf,slen,TXRX_TYPE_CMD);

	/* send data */
	lcount=0;
	bcnt=0;
	count=0;
	checksum=0;
	while(bcnt<nlen) {

		/* uuencode / prepare data bytes */
		uuencode((u8 *)(buf+bcnt),(u8 *)(txrxbuf),
		         (bcnt==nlen-3)?(len%3?len%3:3):3);
		txrxbuf[5]='\r';
		txrxbuf[6]='\n';

		/* checksum */
		checksum+=((u8)buf[bcnt]+(u8)buf[bcnt+1]+(u8)buf[bcnt+2]);

		/* send a data line */
		txrx(lpc,txrxbuf,7,TXRX_TYPE_DATA);

		/* increase counters */
		lcount+=1;
		bcnt+=3;
		count+=3;

		/* checksum */
		if((!(lcount%20))|(bcnt==nlen)) {
			/* send backtick */
			memcpy(txrxbuf,"`\r\n",3);
			//txrx(lpc,txrxbuf,3,TXRX_TYPE_DATA);
			/* send checksum */
			snprintf(txrxbuf,BUFSIZE,"%d\r\n",checksum);
			slen=strlen(txrxbuf);
			txrx(lpc,txrxbuf,slen,TXRX_TYPE_CKSM);
			if(!strncmp(txrxbuf,"RESE",4)) {
				read(lpc->sfd,txrxbuf+4,4);
				printf("ram write: resending ...\n");
				bcnt-=count;
			}
			if(strncmp(txrxbuf,"OK\r\n",4)) {
				printf("ram write: bad response\n");
				return -1;
			}
			/* reset checksum & counter */
			checksum=0;
			count=0;
		}

	}

	return 0;
}

int firmware_to_ram(t_lpc *lpc) {

	char buf[BUFSIZE];
	u32 addr,len,type;
	int ret,temp;

	/* read a line */
	ret=1;
	while(ret) {
		/* sync line */
		ret=read(lpc->fwfd,buf,1);
		switch(buf[0]) {
			case '\r':
				continue;
			case '\n':
				continue;
			case ':':
				/* start code */
				break;
			default:
				printf("fw to ram: no ihex format\n");
				return -1;
		}
		/* read len */
		ret=read(lpc->fwfd,buf,2);
		sscanf(buf,"%02x",&len);
		if(len%4) {
			printf("fw to ram: len not a multiple of 4\n");
			return -1;
		}
		/* read addr */
		ret=read(lpc->fwfd,buf,4);
		sscanf(buf,"%04x",&addr);
		/* read type */
		ret=read(lpc->fwfd,buf,2);
		sscanf(buf,"%02x",&type);
		/* successfull return if type is end of file */
		if(type==0x01)
			return 0;
		/* read data (and cksum) */
		ret=read(lpc->fwfd,buf,2*(len+1));
		if(ret!=(2*(len+1))) {
			printf("fw to ram: data missing\n");
				return -1;
		}
		for(ret=0;ret<len;ret++) {
			sscanf(buf+2*ret,"%02x",&temp);
			buf[ret]=temp;
		}
		/* act according to type */
		switch(type) {
			//case 0x03:
			//	/* get cs and ip */
			//	break;
			case 0x00:
				write_to_ram(lpc,buf,addr,len);
				break;
			case 0x01:
				write_to_ram(lpc,buf,addr,len);
				break;
			default:
				printf("fw to ram: unknown type %02x\n",type);
				return -1;
		}
	}

	return 0;
}

int main(int argc,char **argv) {

	t_lpc lpc;
	int i;
	u8 buf[BUFSIZE];
	int ret;

	/*
	 * initial ... 
	 */

	memset(&lpc,0,sizeof(t_lpc));
	strncpy(lpc.freq,CRYSTFREQ,7);
	lpc.roff=RAMOFFSET;

	/* parse argv */

	for(i=1;i<argc;i++) {

		if(argv[i][0]!='-') {
			usage();
			return -1;
		}

		switch(argv[i][1]) {
			case 'd':
				strncpy(lpc.sdev,argv[++i],127);
				break;
			case 'f':
				strncpy(lpc.fwfile,argv[++i],127);
				lpc.info|=FIRMWARE;
				break;
			case 'v':
				lpc.info|=VERBOSE;
				break;
			case 'c':
				strncpy(lpc.freq,argv[++i],7);
				break;
			default:
				usage();
				return -1;
		}

	}

	/* open serial port */
	if(open_serial_device(&lpc)<0)
		goto end;

	/* boot loader init */
	printf("boot loader init ...\n");
	if(bl_init(&lpc)<0)
		return -1;

	/* quit if there is no hex file to process */
	if(!(lpc.info&FIRMWARE)) {
		printf("no firmware -> aborting\n");
		goto end;
	}

	/* open firmware file */
	if(open_firmware(&lpc)<0)
		goto end;

	/* parse intel hex file and write to ram */
	printf("write firmware to ram ...\n");
	firmware_to_ram(&lpc);

	/* unlock go cmd */
	printf("unlock go command ...\n");
	unlock_go(&lpc);

	/* go! */
	printf("go ...\n");
	ret=go(&lpc);

	/* tell the user that the error might be due to the jump */
	printf("\n\n");
	if(ret<0)
		printf("the above error might be due to the jump!\n\n");

	/* query user for serial port listening */
	printf("continue listening on serial port? (ctrl+c to quit) [y|n]: ");
	buf[0]=getchar();
	printf("\n");

	if(buf[0]!='y')
		goto end;

	/* continue lsitening on serial port */
	ret=1;
	while(ret) {
		ret=read(lpc.sfd,buf,BUFSIZE);
		printf("\rread %d bytes: ",ret);
		for(i=0;i<ret;i++)
			printf("%02x ",buf[i]);
		printf("\n");
	}

end:
	close(lpc.sfd);
	close(lpc.fwfd);

	return 0;
}

